package arc.ex;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import android.util.Log;

public class ControlModule {
	final static int CM_MASTER_ID	= 0xFF;

	// Command
	final static byte CM_OPEN 		= 0;
	final static byte CM_CLOSE 		= 1;
	final static byte CM_RESET 		= 2;
	final static byte CM_PING		= 3;
	final static byte CM_SET_ID 	= 4;
	final static byte CM_GET_ID		= 5;
	final static byte CM_GET_VER	= 6;
	final static byte CM_BLINK 		= 7;
	final static byte CM_TRANSFER 	= 8;
	final static byte CM_SET_PARAM 	= 9;
	final static byte CM_GET_PARAM 	= 10;

	// Status
	final static byte CM_NORMAL			= 100;
	final static byte CM_NAK		  	= 101;
	final static byte CM_CTRL_TIMEOUT	= 102;

	final static byte CM_COMMAND_ERROR		= 110;
	final static byte CM_COMMAND_UNDEFINED 	= 110;
	final static byte CM_COMMAND_TX_TIMEOUT	= 111;		
	final static byte CM_COMMAND_RX_TIMEOUT	= 112;		
	final static byte CM_COMMAND_SUM_ERROR	= 113;		

	final static byte CM_REPLY_ERROR		= 120;
	final static byte CM_REPLY_NO			= 120;
	final static byte CM_REPLY_TX_TIMEOUT	= 121;
	final static byte CM_REPLY_RX_TIMEOUT	= 122;
	final static byte CM_REPLY_SUM_ERROR	= 123;

	final static byte CM_TRANSFER_ERROR			= (byte)130;
	final static byte CM_TRANSFER_TX_TIMEOUT	= (byte)131;
	final static byte CM_TRANSFER_RX_TIMEOUT	= (byte)132;
	final static byte CM_TRANSFER_SUM_ERROR		= (byte)133;

	// response
	final static byte ACK = 0x55; // 85
	final static byte NAK = (byte)0xAA; // 170
	
	final static byte CM_DEFAULT_BUF_SIZE = 64;
	

	public class CmTrHeader {
		byte	port;
		byte	id;
		short	txSize;
		short	rxSize;
	}
	
	final static int TR_HEADER_SIZE	= 6;
	final static int RX_HEADER_SIZE	= 4; // port + id + rxSize
	
	public class CmParam {
		float dt;			// 䃋[v̎[sec]
		float ctrlTimeout;	// ̃^CAEg[sec] 0Ł
		float comTimeout;	// ʐM̃^CAEg[sec] 0Ł
	}
	
	public class CmException extends Exception{
		public CmException(){
			super();
		}
		public CmException( ControlModule cm, String msg )
		{
			super( msg + String.format(": %s(port %d, id %d)", cm.name, cm.port, cm.id ) );
		}
	}
	
	String name;
	byte port;
	byte id;
	byte status;
	byte version;
	String error;
		
	ControlModule parentModule;
	ByteBuffer txBuffer;
	ByteBuffer rxBuffer;
	int bufferSize;
	int reservedRxSize;
	boolean checkSumEnable;
	boolean opened;
	
	byte command;
	
    int childModuleCount;
    ControlModule[] childModule;
	int totalRxSize;

	
	public ControlModule()
	{
		name = "Control Module";
		parentModule = null;
		status = CM_NORMAL;
		txBuffer = ByteBuffer.allocate(CM_DEFAULT_BUF_SIZE);
		rxBuffer = ByteBuffer.allocate(CM_DEFAULT_BUF_SIZE);
		txBuffer.order(ByteOrder.LITTLE_ENDIAN);
		rxBuffer.order(ByteOrder.LITTLE_ENDIAN);
		bufferSize = CM_DEFAULT_BUF_SIZE;
    	totalRxSize = 0;
    	
        childModuleCount = 0;
    	childModule = new ControlModule[256];
		
		checkSumEnable = false;
		version = 0;
		reservedRxSize = 0;
		opened = false;
		

	}

	String dumpRxData( int size )
	{
		String str;
		int i;
		
		str="";
		for(i=0;i<size;i++){
			str += rxBuffer.get(i) + " ";
		}
		return( str );
	}
	
	void addCheckSum()
	{
		int i, size;
		byte sum=0;
		
		size = txBuffer.position();
		txBuffer.rewind();
		for(i=0;i<size;i++){
			sum += txBuffer.get();
		}
		txBuffer.put( sum );
	}

	boolean checkSum( int rxSize)
	{
		int i;
		byte sum=0;
		
		rxBuffer.rewind();

		for(i=0;i<rxSize;i++){
			sum += rxBuffer.get();
		}
		if( rxBuffer.get() != sum ){
			rxBuffer.rewind();
			return( false );
		}
		rxBuffer.rewind();
		return( true );
	}

	void transceive( int rxSize, boolean rightNow ) throws CmException 
	{
		if( parentModule != null ){
			rxSize++; // add status;
		
			if( checkSumEnable ){
				addCheckSum();
				rxSize++; // +Checksum
			}
			reservedRxSize = rxSize;

			parentModule.transfer( this, txBuffer.position(), rxSize, rightNow );
			txBuffer.clear();
			
			if( status != CM_NORMAL ){
				throw new CmException( this, "Error status =" + status );
			}
		
		}
	}
	
    void transfer( ControlModule requestModule, int txSize, int rxSize, boolean rightNow ) throws CmException
	{
		int size, i;
	
		// Header registration 
		if(childModuleCount ==256 ){
			throw new CmException( this, "Child module count exceed" );
		}
		childModule[ childModuleCount ] = requestModule;
		
		if( childModuleCount == 0 ){
			command = CM_TRANSFER;
			txBuffer.put( command );
			txBuffer.put( (byte)childModuleCount ); // MOɍXV
			
			totalRxSize = 0;
		}
		
		// Check transmit data size
		size = txBuffer.position() + TR_HEADER_SIZE + txSize;
		if( size > bufferSize ){
			throw new CmException( 
					this, 
					"Tx size exceeded by chiled module( port " + requestModule.port + " id " + requestModule.id
					);
		}
	
		// Check receive data size
		size =	totalRxSize + RX_HEADER_SIZE + rxSize;
		if( size > bufferSize ){
			throw new CmException(
					this,
					"Rx size exceeded by chiled module( port " + requestModule.port + " id " + requestModule.id
					);
		}
	
		////////////////////////////////////////////////////////////////////////////////////
		// Copy transmit header
		txBuffer.put( requestModule.port );
		txBuffer.put( requestModule.id );
		txBuffer.putShort( (short)txSize );
		txBuffer.putShort( (short)rxSize );
		
		// Copy transmit data
		txBuffer.put( requestModule.txBuffer.array(), 0, txSize );
		
		// Add RX data size
		totalRxSize += RX_HEADER_SIZE + rxSize; // rx header + rxSize
		childModuleCount++;
		
		//Transceive 
		if( !rightNow ) return;
		
		txBuffer.put( 1, (byte)childModuleCount ); // XV
		
		transceive( totalRxSize, true );

		// Copy received data to childModule buffer
		byte childId, childPort;
		int childRxSize, pos;
		
		for(i=0; i<childModuleCount; i++ ){
			childPort = rxBuffer.get();
			childId = rxBuffer.get();
			childRxSize = rxBuffer.getShort(); // received size
			childModule[i].rxBuffer.clear();
			childModule[i].rxBuffer.put( rxBuffer.array(), rxBuffer.position(), childRxSize );
			childModule[i].receive( childRxSize );
			
			pos = rxBuffer.position() + childRxSize;
			rxBuffer.position( pos );
		}
		childModuleCount = 0;
	}

	void receive( int rxSize ) throws CmException 
	{
		//Log.i("DATK","rx " + rxSize + " bytes: " + dumpRxData( rxSize ));
		
		if( rxSize != reservedRxSize ){
			if( checkSumEnable ){
				rxSize--; // sub checksum;

				if( rxSize == 0 ){
					status = rxBuffer.get();
				}
			}
			status = rxBuffer.get(rxSize-1);
			throw new CmException(
					this, 
					"Rx data size mismatch, required " + reservedRxSize + " actual " + rxSize + " status " + status
					);
		}
		
		if( checkSumEnable ){
			rxSize--; // -CheckSum
		    if(!checkSum( rxSize )){
				status = (byte)CM_REPLY_SUM_ERROR;
				throw new CmException( this, "Checksum error " + status	);
			}
		}
		
		status = rxBuffer.get(rxSize-1);
		if( status!=CM_NORMAL ){
			throw new CmException( this, "Error status " + status);
		}
		
		rxBuffer.rewind();
	}

	void open( int id ) throws CmException 
	{
		parentModule = null;
		opened = true;
		command = (byte)CM_OPEN;
		
		txBuffer.clear();
		rxBuffer.clear();
		childModuleCount = 0;
	}
	
	
	void open( ControlModule parent, int parentPort, int targetId ) throws CmException 
	{
		byte b;
		
		parentModule = parent;
		port = (byte)parentPort;
		id   = (byte)targetId;
		
		txBuffer.clear();
		rxBuffer.clear();
		childModuleCount = 0;
		
		command = CM_OPEN;
		txBuffer.put(command);
		
		transceive( 1, true );
		
		b = rxBuffer.get();
		if( b != CM_OPEN  ){
			throw new CmException( this, "Illegal return value " + b );
		}

		opened = true;
	}

	void close() throws CmException 
	{
		byte b;
		
		if( !opened ) return;
	
		command = CM_CLOSE;
		txBuffer.put( command );
		
		transceive( 1, true );
		
		b = rxBuffer.get();
		if( b != CM_CLOSE  ){
			throw new CmException( this, "Illegal return value " + b );
		}
		opened = false;

	}
	
	boolean isOpen()
	{
		return(opened);
	}

	
	void reset() throws CmException 
	{
		command = CM_RESET;
		txBuffer.put(command);
		
		transceive( 0, true );

		try {
			Thread.sleep(1000); //foCXA܂ŎԒ
		}
		catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	void ping() throws CmException 
	{
		byte b;
		
		command = CM_PING;
		txBuffer.put(command);
		
		transceive( 0, true );
		b = rxBuffer.get();
		if( b != CM_PING  ){
			throw new CmException( this, "Return value mismatch " + b );
		}
	}

	void setID( byte newId ) throws CmException 
	{
		command = CM_SET_ID;
		txBuffer.put(command);
		txBuffer.put(id);
		
		transceive( 0, true );
		
		id = newId;
	}

	byte getID() throws CmException 
	{
		command = CM_GET_ID;
		txBuffer.put(command);
		
		transceive( 1, true );
		
		id = rxBuffer.get();
		return( id );
	}

	byte getVersion() throws CmException 
	{
		command = CM_GET_VER;
		txBuffer.put(command);
		
		transceive( 1, true );

		version = rxBuffer.get();
		return( version ); 
	}

	void blink( int num ) throws CmException 	
	{
//		throw new CmException( this, "Blink! " );

		command = CM_BLINK;
		txBuffer.put(command);
		txBuffer.put((byte)num);
		
		transceive( 0, true );
	}

	void setParam( CmParam param ) throws CmException 
	{
		command = CM_SET_PARAM;
		txBuffer.put(command);
		txBuffer.putFloat( param.dt );
		txBuffer.putFloat( param.ctrlTimeout );
		txBuffer.putFloat( param.comTimeout );
		
		transceive( 0, true );
	}

	void getParam( CmParam param  ) throws CmException
	{
		command = CM_GET_PARAM;
		txBuffer.put(command);
		
		transceive( 12, true );
		
		param.dt = rxBuffer.getFloat();
		param.ctrlTimeout = rxBuffer.getFloat();
		param.comTimeout = rxBuffer.getFloat();
	}
	
}

